@@ -35,6 +35,8 @@
 #include "lib/pthreadpool/pthreadpool_tevent.h"
 #include "librpc/gen_ndr/ndr_ioctl.h"
 #include "offload_token.h"
+#include "util_reparse.h"
+#include "lib/util/string_wrappers.h"
 
 #undef DBGC_CLASS
 #define DBGC_CLASS DBGC_VFS
@@ -47,7 +49,8 @@
 
 static int vfswrap_connect(vfs_handle_struct *handle, const char *service, const char *user)
 {
-    return 0;    /* Return >= 0 for success */
+	handle->conn->have_proc_fds = sys_have_proc_fds();
+	return 0;    /* Return >= 0 for success */
 }
 
 static void vfswrap_disconnect(vfs_handle_struct *handle)
@@ -131,8 +134,12 @@ static uint32_t vfswrap_fs_capabilities(struct vfs_handle_struct *handle,
 	struct vfs_statvfs_struct statbuf;
 	int ret;
 
-	smb_fname_cpath = synthetic_smb_fname(talloc_tos(), conn->connectpath,
-					      NULL, NULL, 0);
+	smb_fname_cpath = synthetic_smb_fname(talloc_tos(),
+					      conn->connectpath,
+					      NULL,
+					      NULL,
+					      0,
+					      0);
 	if (smb_fname_cpath == NULL) {
 		return caps;
 	}
@@ -390,7 +397,7 @@ static NTSTATUS vfswrap_create_dfs_pathat(struct vfs_handle_struct *handle,
 	}
 
 	ret = symlinkat(msdfs_link,
-			dirfsp->fh->fd,
+			fsp_get_io_fd(dirfsp),
 			smb_fname->base_name);
 	if (ret == 0) {
 		status = NT_STATUS_OK;
@@ -416,7 +423,7 @@ static NTSTATUS vfswrap_create_dfs_pathat(struct vfs_handle_struct *handle,
 static NTSTATUS vfswrap_read_dfs_pathat(struct vfs_handle_struct *handle,
 				TALLOC_CTX *mem_ctx,
 				struct files_struct *dirfsp,
-				const struct smb_filename *smb_fname,
+				struct smb_filename *smb_fname,
 				struct referral **ppreflist,
 				size_t *preferral_count)
 {
@@ -430,9 +437,15 @@ static NTSTATUS vfswrap_read_dfs_pathat(struct vfs_handle_struct *handle,
 #else
 	char link_target_buf[7];
 #endif
+	int ret;
 
 	SMB_ASSERT(dirfsp == dirfsp->conn->cwd_fsp);
 
+	if (is_named_stream(smb_fname)) {
+		status = NT_STATUS_OBJECT_NAME_NOT_FOUND;
+		goto err;
+	}
+
 	if (ppreflist == NULL && preferral_count == NULL) {
 		/*
 		 * We're only checking if this is a DFS
@@ -448,7 +461,7 @@ static NTSTATUS vfswrap_read_dfs_pathat(struct vfs_handle_struct *handle,
 		}
 	}
 
-	referral_len = readlinkat(dirfsp->fh->fd,
+	referral_len = readlinkat(fsp_get_io_fd(dirfsp),
 				smb_fname->base_name,
 				link_target,
 				bufsize - 1);
@@ -481,6 +494,14 @@ static NTSTATUS vfswrap_read_dfs_pathat(struct vfs_handle_struct *handle,
 		goto err;
 	}
 
+	ret = sys_lstat(smb_fname->base_name,
+			&smb_fname->st,
+			lp_fake_directory_create_times(SNUM(handle->conn)));
+	if (ret < 0) {
+		status = map_nt_error_from_unix(errno);
+		goto err;
+	}
+
 	if (ppreflist == NULL && preferral_count == NULL) {
 		/* Early return for checking if this is a DFS link. */
 		return NT_STATUS_OK;
@@ -535,19 +556,6 @@ static NTSTATUS vfswrap_snap_delete(struct vfs_handle_struct *handle,
 
 /* Directory operations */
 
-static DIR *vfswrap_opendir(vfs_handle_struct *handle,
-				const struct smb_filename *smb_fname,
-				const char *mask,
-				uint32_t attr)
-{
-	DIR *result;
-
-	START_PROFILE(syscall_opendir);
-	result = opendir(smb_fname->base_name);
-	END_PROFILE(syscall_opendir);
-	return result;
-}
-
 static DIR *vfswrap_fdopendir(vfs_handle_struct *handle,
 			files_struct *fsp,
 			const char *mask,
@@ -556,51 +564,72 @@ static DIR *vfswrap_fdopendir(vfs_handle_struct *handle,
 	DIR *result;
 
 	START_PROFILE(syscall_fdopendir);
-	result = sys_fdopendir(fsp->fh->fd);
+	result = sys_fdopendir(fsp_get_io_fd(fsp));
 	END_PROFILE(syscall_fdopendir);
 	return result;
 }
 
 
 static struct dirent *vfswrap_readdir(vfs_handle_struct *handle,
-				          DIR *dirp,
-					  SMB_STRUCT_STAT *sbuf)
+				      struct files_struct *dirfsp,
+				      DIR *dirp,
+				      SMB_STRUCT_STAT *sbuf)
 {
 	struct dirent *result;
+	bool do_stat = false;
+	bool fake_ctime = lp_fake_directory_create_times(SNUM(handle->conn));
+	int flags = AT_SYMLINK_NOFOLLOW;
+	struct stat st;
+	int ret;
 
 	START_PROFILE(syscall_readdir);
+
+#if defined(HAVE_DIRFD) && defined(HAVE_FSTATAT)
+	do_stat = true;
+#endif
+
 	result = readdir(dirp);
 	END_PROFILE(syscall_readdir);
-	if (sbuf) {
-		/* Default Posix readdir() does not give us stat info.
-		 * Set to invalid to indicate we didn't return this info. */
-		SET_STAT_INVALID(*sbuf);
-#if defined(HAVE_DIRFD) && defined(HAVE_FSTATAT)
-		if (result != NULL) {
-			/* See if we can efficiently return this. */
-			struct stat st;
-			int flags = AT_SYMLINK_NOFOLLOW;
-			int ret = fstatat(dirfd(dirp),
-					result->d_name,
-					&st,
-					flags);
-			/*
-			 * As this is an optimization,
-			 * ignore it if we stat'ed a
-			 * symlink. Make the caller
-			 * do it again as we don't
-			 * know if they wanted the link
-			 * info, or its target info.
-			 */
-			if ((ret == 0) && (!S_ISLNK(st.st_mode))) {
-				init_stat_ex_from_stat(sbuf,
-					&st,
-					lp_fake_directory_create_times(
-						SNUM(handle->conn)));
-			}
-		}
-#endif
+
+	if (sbuf == NULL) {
+		return result;
 	}
+	if (result == NULL) {
+		return NULL;
+	}
+
+	/*
+	 * Default Posix readdir() does not give us stat info.
+	 * Set to invalid to indicate we didn't return this info.
+	 */
+	SET_STAT_INVALID(*sbuf);
+
+	/* See if we can efficiently return this. */
+	if (!do_stat) {
+		return result;
+	}
+
+	ret = fstatat(dirfd(dirp),
+		      result->d_name,
+		      &st,
+		      flags);
+	if (ret != 0) {
+		return result;
+	}
+
+	/*
+	 * As this is an optimization, ignore it if we stat'ed a
+	 * symlink for non-POSIX context. Make the caller do it again
+	 * as we don't know if they wanted the link info, or its
+	 * target info.
+	 */
+	if (S_ISLNK(st.st_mode) &&
+	    !(dirfsp->fsp_name->flags & SMB_FILENAME_POSIX_PATH))
+	{
+		return result;
+	}
+	init_stat_ex_from_stat(sbuf, &st, fake_ctime);
+
 	return result;
 }
 
@@ -641,22 +670,10 @@ static int vfswrap_mkdirat(vfs_handle_struct *handle,
 			mode_t mode)
 {
 	int result;
-	const char *path = smb_fname->base_name;
-	char *parent = NULL;
 
 	START_PROFILE(syscall_mkdirat);
 
-	SMB_ASSERT(dirfsp == dirfsp->conn->cwd_fsp);
-
-	if (lp_inherit_acls(SNUM(handle->conn))
-	    && parent_dirname(talloc_tos(), path, &parent, NULL)
-	    && directory_has_default_acl(handle->conn, parent)) {
-		mode = (0777 & lp_directory_mask(SNUM(handle->conn)));
-	}
-
-	TALLOC_FREE(parent);
-
-	result = mkdirat(dirfsp->fh->fd, path, mode);
+	result = mkdirat(fsp_get_pathref_fd(dirfsp), smb_fname->base_name, mode);
 
 	END_PROFILE(syscall_mkdirat);
 	return result;
@@ -674,28 +691,54 @@ static int vfswrap_closedir(vfs_handle_struct *handle, DIR *dirp)
 
 /* File operations */
 
-static int vfswrap_open(vfs_handle_struct *handle,
-			struct smb_filename *smb_fname,
-			files_struct *fsp, int flags, mode_t mode)
+static int vfswrap_openat(vfs_handle_struct *handle,
+			  const struct files_struct *dirfsp,
+			  const struct smb_filename *smb_fname,
+			  files_struct *fsp,
+			  int flags,
+			  mode_t mode)
 {
-	int result = -1;
+	bool have_opath = false;
+	bool became_root = false;
+	int result;
 
-	START_PROFILE(syscall_open);
+	START_PROFILE(syscall_openat);
 
 	if (is_named_stream(smb_fname)) {
 		errno = ENOENT;
+		result = -1;
 		goto out;
 	}
 
-	result = open(smb_fname->base_name, flags, mode);
- out:
-	END_PROFILE(syscall_open);
+#ifdef O_PATH
+	have_opath = true;
+	if (fsp->fsp_flags.is_pathref) {
+		flags |= O_PATH;
+	}
+#endif
+
+	if (fsp->fsp_flags.is_pathref && !have_opath) {
+		become_root();
+		became_root = true;
+	}
+
+	result = openat(fsp_get_pathref_fd(dirfsp),
+			smb_fname->base_name,
+			flags,
+			mode);
+
+	if (became_root) {
+		unbecome_root();
+	}
+
+	fsp->fsp_flags.have_proc_fds = fsp->conn->have_proc_fds;
+
+out:
+	END_PROFILE(syscall_openat);
 	return result;
 }
-
 static NTSTATUS vfswrap_create_file(vfs_handle_struct *handle,
 				    struct smb_request *req,
-				    uint16_t root_dir_fid,
 				    struct smb_filename *smb_fname,
 				    uint32_t access_mask,
 				    uint32_t share_access,
@@ -713,7 +756,7 @@ static NTSTATUS vfswrap_create_file(vfs_handle_struct *handle,
 				    const struct smb2_create_blobs *in_context_blobs,
 				    struct smb2_create_blobs *out_context_blobs)
 {
-	return create_file_default(handle->conn, req, root_dir_fid, smb_fname,
+	return create_file_default(handle->conn, req, smb_fname,
 				   access_mask, share_access,
 				   create_disposition, create_options,
 				   file_attributes, oplock_request, lease,
@@ -739,13 +782,13 @@ static ssize_t vfswrap_pread(vfs_handle_struct *handle, files_struct *fsp, void
 
 #if defined(HAVE_PREAD) || defined(HAVE_PREAD64)
 	START_PROFILE_BYTES(syscall_pread, n);
-	result = sys_pread_full(fsp->fh->fd, data, n, offset);
+	result = sys_pread_full(fsp_get_io_fd(fsp), data, n, offset);
 	END_PROFILE_BYTES(syscall_pread);
 
 	if (result == -1 && errno == ESPIPE) {
 		/* Maintain the fiction that pipes can be seeked (sought?) on. */
-		result = sys_read(fsp->fh->fd, data, n);
-		fsp->fh->pos = 0;
+		result = sys_read(fsp_get_io_fd(fsp), data, n);
+		fh_set_pos(fsp->fh, 0);
 	}
 
 #else /* HAVE_PREAD */
@@ -763,12 +806,12 @@ static ssize_t vfswrap_pwrite(vfs_handle_struct *handle, files_struct *fsp, cons
 
 #if defined(HAVE_PWRITE) || defined(HAVE_PRWITE64)
 	START_PROFILE_BYTES(syscall_pwrite, n);
-	result = sys_pwrite_full(fsp->fh->fd, data, n, offset);
+	result = sys_pwrite_full(fsp_get_io_fd(fsp), data, n, offset);
 	END_PROFILE_BYTES(syscall_pwrite);
 
 	if (result == -1 && errno == ESPIPE) {
 		/* Maintain the fiction that pipes can be sought on. */
-		result = sys_write(fsp->fh->fd, data, n);
+		result = sys_write(fsp_get_io_fd(fsp), data, n);
 	}
 
 #else /* HAVE_PWRITE */
@@ -810,7 +853,7 @@ static struct tevent_req *vfswrap_pread_send(struct vfs_handle_struct *handle,
 	}
 
 	state->ret = -1;
-	state->fd = fsp->fh->fd;
+	state->fd = fsp_get_io_fd(fsp);
 	state->buf = data;
 	state->count = n;
 	state->offset = offset;
@@ -938,7 +981,7 @@ static struct tevent_req *vfswrap_pwrite_send(struct vfs_handle_struct *handle,
 	}
 
 	state->ret = -1;
-	state->fd = fsp->fh->fd;
+	state->fd = fsp_get_io_fd(fsp);
 	state->buf = data;
 	state->count = n;
 	state->offset = offset;
@@ -1061,7 +1104,7 @@ static struct tevent_req *vfswrap_fsync_send(struct vfs_handle_struct *handle,
 	}
 
 	state->ret = -1;
-	state->fd = fsp->fh->fd;
+	state->fd = fsp_get_io_fd(fsp);
 
 	SMBPROFILE_BYTES_ASYNC_START(syscall_asys_fsync, profile_p,
 				     state->profile_bytes, 0);
@@ -1159,7 +1202,7 @@ static off_t vfswrap_lseek(vfs_handle_struct *handle, files_struct *fsp, off_t o
 
 	START_PROFILE(syscall_lseek);
 
-	result = lseek(fsp->fh->fd, offset, whence);
+	result = lseek(fsp_get_io_fd(fsp), offset, whence);
 	/*
 	 * We want to maintain the fiction that we can seek
 	 * on a fifo for file system purposes. This allows
@@ -1182,7 +1225,7 @@ static ssize_t vfswrap_sendfile(vfs_handle_struct *handle, int tofd, files_struc
 	ssize_t result;
 
 	START_PROFILE_BYTES(syscall_sendfile, n);
-	result = sys_sendfile(tofd, fromfsp->fh->fd, hdr, offset, n);
+	result = sys_sendfile(tofd, fsp_get_io_fd(fromfsp), hdr, offset, n);
 	END_PROFILE_BYTES(syscall_sendfile);
 	return result;
 }
@@ -1196,7 +1239,7 @@ static ssize_t vfswrap_recvfile(vfs_handle_struct *handle,
 	ssize_t result;
 
 	START_PROFILE_BYTES(syscall_recvfile, n);
-	result = sys_recvfile(fromfd, tofsp->fh->fd, offset, n);
+	result = sys_recvfile(fromfd, fsp_get_io_fd(tofsp), offset, n);
 	END_PROFILE_BYTES(syscall_recvfile);
 	return result;
 }
@@ -1216,9 +1259,9 @@ static int vfswrap_renameat(vfs_handle_struct *handle,
 		goto out;
 	}
 
-	result = renameat(srcfsp->fh->fd,
+	result = renameat(fsp_get_pathref_fd(srcfsp),
 			smb_fname_src->base_name,
-			dstfsp->fh->fd,
+			fsp_get_pathref_fd(dstfsp),
 			smb_fname_dst->base_name);
 
  out:
@@ -1250,7 +1293,7 @@ static int vfswrap_fstat(vfs_handle_struct *handle, files_struct *fsp, SMB_STRUC
 	int result;
 
 	START_PROFILE(syscall_fstat);
-	result = sys_fstat(fsp->fh->fd,
+	result = sys_fstat(fsp_get_pathref_fd(fsp),
 			   sbuf, lp_fake_directory_create_times(SNUM(handle->conn)));
 	END_PROFILE(syscall_fstat);
 	return result;
@@ -1354,18 +1397,21 @@ static NTSTATUS vfswrap_fsctl(struct vfs_handle_struct *handle,
 
 	case FSCTL_GET_REPARSE_POINT:
 	{
-		/* Fail it with STATUS_NOT_A_REPARSE_POINT */
-		DEBUG(10, ("FSCTL_GET_REPARSE_POINT: called on %s. "
-			   "Status: NOT_IMPLEMENTED\n", fsp_fnum_dbg(fsp)));
-		return NT_STATUS_NOT_A_REPARSE_POINT;
+		status = fsctl_get_reparse_point(
+			fsp, ctx, out_data, max_out_len, out_len);
+		return status;
 	}
 
 	case FSCTL_SET_REPARSE_POINT:
 	{
-		/* Fail it with STATUS_NOT_A_REPARSE_POINT */
-		DEBUG(10, ("FSCTL_SET_REPARSE_POINT: called on %s. "
-			   "Status: NOT_IMPLEMENTED\n", fsp_fnum_dbg(fsp)));
-		return NT_STATUS_NOT_A_REPARSE_POINT;
+		status = fsctl_set_reparse_point(fsp, ctx, _in_data, in_len);
+		return status;
+	}
+
+	case FSCTL_DELETE_REPARSE_POINT:
+	{
+		status = fsctl_del_reparse_point(fsp, ctx, _in_data, in_len);
+		return status;
 	}
 
 	case FSCTL_GET_SHADOW_COPY_DATA:
@@ -1641,20 +1687,6 @@ static NTSTATUS vfswrap_fsctl(struct vfs_handle_struct *handle,
 static bool vfswrap_is_offline(struct connection_struct *conn,
 			       const struct smb_filename *fname);
 
-static NTSTATUS vfswrap_get_dos_attributes(struct vfs_handle_struct *handle,
-					   struct smb_filename *smb_fname,
-					   uint32_t *dosmode)
-{
-	bool offline;
-
-	offline = vfswrap_is_offline(handle->conn, smb_fname);
-	if (offline) {
-		*dosmode |= FILE_ATTRIBUTE_OFFLINE;
-	}
-
-	return get_ea_dos_attribute(handle->conn, smb_fname, dosmode);
-}
-
 struct vfswrap_get_dos_attributes_state {
 	struct vfs_aio_state aio_state;
 	connection_struct *conn;
@@ -1788,6 +1820,7 @@ static void vfswrap_get_dos_attributes_getxattr_done(struct tevent_req *subreq)
 		.base_name = path,
 		.st = state->smb_fname->st,
 		.flags = state->smb_fname->flags,
+		.twrp = state->smb_fname->twrp,
 	};
 
 	offline = vfswrap_is_offline(state->conn, &smb_fname);
@@ -1831,7 +1864,7 @@ static NTSTATUS vfswrap_fget_dos_attributes(struct vfs_handle_struct *handle,
 		*dosmode |= FILE_ATTRIBUTE_OFFLINE;
 	}
 
-	return get_ea_dos_attribute(handle->conn, fsp->fsp_name, dosmode);
+	return fget_ea_dos_attribute(fsp, dosmode);
 }
 
 static NTSTATUS vfswrap_set_dos_attributes(struct vfs_handle_struct *handle,
@@ -2288,10 +2321,9 @@ static NTSTATUS vfswrap_offload_write_recv(struct vfs_handle_struct *handle,
 	return NT_STATUS_OK;
 }
 
-static NTSTATUS vfswrap_get_compression(struct vfs_handle_struct *handle,
+static NTSTATUS vfswrap_fget_compression(struct vfs_handle_struct *handle,
 					TALLOC_CTX *mem_ctx,
 					struct files_struct *fsp,
-					struct smb_filename *smb_fname,
 					uint16_t *_compression_fmt)
 {
 	return NT_STATUS_INVALID_DEVICE_REQUEST;
@@ -2378,7 +2410,7 @@ static int vfswrap_unlinkat(vfs_handle_struct *handle,
 		errno = ENOENT;
 		goto out;
 	}
-	result = unlinkat(dirfsp->fh->fd,
+	result = unlinkat(fsp_get_pathref_fd(dirfsp),
 			smb_fname->base_name,
 			flags);
 
@@ -2405,7 +2437,7 @@ static int vfswrap_fchmod(vfs_handle_struct *handle, files_struct *fsp, mode_t m
 
 	START_PROFILE(syscall_fchmod);
 #if defined(HAVE_FCHMOD)
-	result = fchmod(fsp->fh->fd, mode);
+	result = fchmod(fsp_get_io_fd(fsp), mode);
 #else
 	result = -1;
 	errno = ENOSYS;
@@ -2421,7 +2453,7 @@ static int vfswrap_fchown(vfs_handle_struct *handle, files_struct *fsp, uid_t ui
 	int result;
 
 	START_PROFILE(syscall_fchown);
-	result = fchown(fsp->fh->fd, uid, gid);
+	result = fchown(fsp_get_io_fd(fsp), uid, gid);
 	END_PROFILE(syscall_fchown);
 	return result;
 #else
@@ -2471,6 +2503,7 @@ static struct smb_filename *vfswrap_getwd(vfs_handle_struct *handle,
 				result,
 				NULL,
 				NULL,
+				0,
 				0);
 	/*
 	 * sys_getwd() *always* returns malloced memory.
@@ -2606,7 +2639,7 @@ static int strict_allocate_ftruncate(vfs_handle_struct *handle, files_struct *fs
 
 	/* Shrink - just ftruncate. */
 	if (pst->st_ex_size > len)
-		return ftruncate(fsp->fh->fd, len);
+		return ftruncate(fsp_get_io_fd(fsp), len);
 
 	space_to_write = len - pst->st_ex_size;
 
@@ -2653,7 +2686,7 @@ static int vfswrap_ftruncate(vfs_handle_struct *handle, files_struct *fsp, off_t
 
 	START_PROFILE(syscall_ftruncate);
 
-	if (lp_strict_allocate(SNUM(fsp->conn)) && !fsp->is_sparse) {
+	if (lp_strict_allocate(SNUM(fsp->conn)) && !fsp->fsp_flags.is_sparse) {
 		result = strict_allocate_ftruncate(handle, fsp, len);
 		END_PROFILE(syscall_ftruncate);
 		return result;
@@ -2665,7 +2698,7 @@ static int vfswrap_ftruncate(vfs_handle_struct *handle, files_struct *fsp, off_t
 	   expansion and some that don't! On Linux fat can't do
 	   ftruncate extend but ext2 can. */
 
-	result = ftruncate(fsp->fh->fd, len);
+	result = ftruncate(fsp_get_io_fd(fsp), len);
 
 	/* According to W. R. Stevens advanced UNIX prog. Pure 4.3 BSD cannot
 	   extend a file with ftruncate. Provide alternate implementation
@@ -2726,7 +2759,7 @@ static int vfswrap_fallocate(vfs_handle_struct *handle,
 
 	START_PROFILE(syscall_fallocate);
 	if (mode == 0) {
-		result = sys_posix_fallocate(fsp->fh->fd, offset, len);
+		result = sys_posix_fallocate(fsp_get_io_fd(fsp), offset, len);
 		/*
 		 * posix_fallocate returns 0 on success, errno on error
 		 * and doesn't set errno. Make it behave like fallocate()
@@ -2738,7 +2771,7 @@ static int vfswrap_fallocate(vfs_handle_struct *handle,
 		}
 	} else {
 		/* sys_fallocate handles filtering of unsupported mode flags */
-		result = sys_fallocate(fsp->fh->fd, mode, offset, len);
+		result = sys_fallocate(fsp_get_io_fd(fsp), mode, offset, len);
 	}
 	END_PROFILE(syscall_fallocate);
 	return result;
@@ -2750,11 +2783,11 @@ static bool vfswrap_lock(vfs_handle_struct *handle, files_struct *fsp, int op, o
 
 	START_PROFILE(syscall_fcntl_lock);
 
-	if (fsp->use_ofd_locks) {
+	if (fsp->fsp_flags.use_ofd_locks) {
 		op = map_process_lock_to_ofd_lock(op);
 	}
 
-	result =  fcntl_lock(fsp->fh->fd, op, offset, count, type);
+	result =  fcntl_lock(fsp_get_io_fd(fsp), op, offset, count, type);
 	END_PROFILE(syscall_fcntl_lock);
 	return result;
 }
@@ -2763,7 +2796,7 @@ static int vfswrap_kernel_flock(vfs_handle_struct *handle, files_struct *fsp,
 				uint32_t share_access, uint32_t access_mask)
 {
 	START_PROFILE(syscall_kernel_flock);
-	kernel_flock(fsp->fh->fd, share_access, access_mask);
+	kernel_flock(fsp_get_io_fd(fsp), share_access, access_mask);
 	END_PROFILE(syscall_kernel_flock);
 	return 0;
 }
@@ -2800,11 +2833,11 @@ static int vfswrap_fcntl(vfs_handle_struct *handle, files_struct *fsp, int cmd,
 	case F_SET_FILE_RW_HINT:
 #endif
 		argp = va_arg(dup_cmd_arg, void *);
-		result = sys_fcntl_ptr(fsp->fh->fd, cmd, argp);
+		result = sys_fcntl_ptr(fsp_get_io_fd(fsp), cmd, argp);
 		break;
 	default:
 		val = va_arg(dup_cmd_arg, int);
-		result = sys_fcntl_int(fsp->fh->fd, cmd, val);
+		result = sys_fcntl_int(fsp_get_io_fd(fsp), cmd, val);
 	}
 
 	va_end(dup_cmd_arg);
@@ -2820,11 +2853,11 @@ static bool vfswrap_getlock(vfs_handle_struct *handle, files_struct *fsp, off_t
 
 	START_PROFILE(syscall_fcntl_getlock);
 
-	if (fsp->use_ofd_locks) {
+	if (fsp->fsp_flags.use_ofd_locks) {
 		op = map_process_lock_to_ofd_lock(op);
 	}
 
-	result = fcntl_getlock(fsp->fh->fd, op, poffset, pcount, ptype, ppid);
+	result = fcntl_getlock(fsp_get_io_fd(fsp), op, poffset, pcount, ptype, ppid);
 	END_PROFILE(syscall_fcntl_getlock);
 	return result;
 }
@@ -2837,7 +2870,7 @@ static int vfswrap_linux_setlease(vfs_handle_struct *handle, files_struct *fsp,
 	START_PROFILE(syscall_linux_setlease);
 
 #ifdef HAVE_KERNEL_OPLOCKS_LINUX
-	result = linux_setlease(fsp->fh->fd, leasetype);
+	result = linux_setlease(fsp_get_io_fd(fsp), leasetype);
 #else
 	errno = ENOSYS;
 #endif
@@ -2846,7 +2879,7 @@ static int vfswrap_linux_setlease(vfs_handle_struct *handle, files_struct *fsp,
 }
 
 static int vfswrap_symlinkat(vfs_handle_struct *handle,
-			const char *link_target,
+			const struct smb_filename *link_target,
 			struct files_struct *dirfsp,
 			const struct smb_filename *new_smb_fname)
 {
@@ -2854,17 +2887,15 @@ static int vfswrap_symlinkat(vfs_handle_struct *handle,
 
 	START_PROFILE(syscall_symlinkat);
 
-	SMB_ASSERT(dirfsp == dirfsp->conn->cwd_fsp);
-
-	result = symlinkat(link_target,
-			dirfsp->fh->fd,
+	result = symlinkat(link_target->base_name,
+			fsp_get_pathref_fd(dirfsp),
 			new_smb_fname->base_name);
 	END_PROFILE(syscall_symlinkat);
 	return result;
 }
 
 static int vfswrap_readlinkat(vfs_handle_struct *handle,
-			files_struct *dirfsp,
+			const struct files_struct *dirfsp,
 			const struct smb_filename *smb_fname,
 			char *buf,
 			size_t bufsiz)
@@ -2873,9 +2904,7 @@ static int vfswrap_readlinkat(vfs_handle_struct *handle,
 
 	START_PROFILE(syscall_readlinkat);
 
-	SMB_ASSERT(dirfsp == dirfsp->conn->cwd_fsp);
-
-	result = readlinkat(dirfsp->fh->fd,
+	result = readlinkat(fsp_get_pathref_fd(dirfsp),
 			smb_fname->base_name,
 			buf,
 			bufsiz);
@@ -2895,12 +2924,9 @@ static int vfswrap_linkat(vfs_handle_struct *handle,
 
 	START_PROFILE(syscall_linkat);
 
-	SMB_ASSERT(srcfsp == srcfsp->conn->cwd_fsp);
-	SMB_ASSERT(dstfsp == dstfsp->conn->cwd_fsp);
-
-	result = linkat(srcfsp->fh->fd,
+	result = linkat(fsp_get_pathref_fd(srcfsp),
 			old_smb_fname->base_name,
-			dstfsp->fh->fd,
+			fsp_get_pathref_fd(dstfsp),
 			new_smb_fname->base_name,
 			flags);
 
@@ -2918,9 +2944,7 @@ static int vfswrap_mknodat(vfs_handle_struct *handle,
 
 	START_PROFILE(syscall_mknodat);
 
-	SMB_ASSERT(dirfsp == dirfsp->conn->cwd_fsp);
-
-	result = sys_mknodat(dirfsp->fh->fd,
+	result = sys_mknodat(fsp_get_pathref_fd(dirfsp),
 			smb_fname->base_name,
 			mode,
 			dev);
@@ -2940,7 +2964,12 @@ static struct smb_filename *vfswrap_realpath(vfs_handle_struct *handle,
 	result = sys_realpath(smb_fname->base_name);
 	END_PROFILE(syscall_realpath);
 	if (result) {
-		result_fname = synthetic_smb_fname(ctx, result, NULL, NULL, 0);
+		result_fname = synthetic_smb_fname(ctx,
+						   result,
+						   NULL,
+						   NULL,
+						   0,
+						   0);
 		SAFE_FREE(result);
 	}
 	return result_fname;
@@ -3007,30 +3036,27 @@ static NTSTATUS vfswrap_streaminfo(vfs_handle_struct *handle,
 	struct stream_struct *tmp_streams = NULL;
 	int ret;
 
-	if ((fsp != NULL) && (fsp->is_directory)) {
+	if ((fsp != NULL) && (fsp->fsp_flags.is_directory)) {
 		/*
 		 * No default streams on directories
 		 */
 		goto done;
 	}
 
-	if ((fsp != NULL) && (fsp->fh->fd != -1)) {
+	if ((fsp != NULL) && (fsp_get_pathref_fd(fsp) != -1)) {
 		ret = SMB_VFS_FSTAT(fsp, &sbuf);
 	}
 	else {
-		struct smb_filename smb_fname_cp;
+		struct smb_filename *smb_fname_cp = NULL;
 
-		ZERO_STRUCT(smb_fname_cp);
-		smb_fname_cp.base_name = discard_const_p(char,
-					smb_fname->base_name);
-		smb_fname_cp.flags = smb_fname->flags;
-
-		if (smb_fname_cp.flags & SMB_FILENAME_POSIX_PATH) {
-			ret = SMB_VFS_LSTAT(handle->conn, &smb_fname_cp);
-		} else {
-			ret = SMB_VFS_STAT(handle->conn, &smb_fname_cp);
+		smb_fname_cp = cp_smb_filename_nostream(talloc_tos(), smb_fname);
+		if (smb_fname_cp == NULL) {
+			return NT_STATUS_NO_MEMORY;
 		}
-		sbuf = smb_fname_cp.st;
+
+		ret = vfs_stat(handle->conn, smb_fname_cp);
+		sbuf = smb_fname_cp->st;
+		TALLOC_FREE(smb_fname_cp);
 	}
 
 	if (ret == -1) {
@@ -3060,7 +3086,7 @@ static NTSTATUS vfswrap_streaminfo(vfs_handle_struct *handle,
 }
 
 static int vfswrap_get_real_filename(struct vfs_handle_struct *handle,
-				     const char *path,
+				     const struct smb_filename *path,
 				     const char *name,
 				     TALLOC_CTX *mem_ctx,
 				     char **found_name)
@@ -3125,21 +3151,25 @@ static NTSTATUS vfswrap_fget_nt_acl(vfs_handle_struct *handle,
 	return result;
 }
 
-static NTSTATUS vfswrap_get_nt_acl(vfs_handle_struct *handle,
-				   const struct smb_filename *smb_fname,
-				   uint32_t security_info,
-				   TALLOC_CTX *mem_ctx,
-				   struct security_descriptor **ppdesc)
+static NTSTATUS vfswrap_get_nt_acl_at(vfs_handle_struct *handle,
+			struct files_struct *dirfsp,
+			const struct smb_filename *smb_fname,
+			uint32_t security_info,
+			TALLOC_CTX *mem_ctx,
+			struct security_descriptor **ppdesc)
 {
 	NTSTATUS result;
 
-	START_PROFILE(get_nt_acl);
+	START_PROFILE(get_nt_acl_at);
+
+	SMB_ASSERT(dirfsp == dirfsp->conn->cwd_fsp);
+
 	result = posix_get_nt_acl(handle->conn,
 				smb_fname,
 				security_info,
 				mem_ctx,
 				ppdesc);
-	END_PROFILE(get_nt_acl);
+	END_PROFILE(get_nt_acl_at);
 	return result;
 }
 
@@ -3177,17 +3207,45 @@ static SMB_ACL_T vfswrap_sys_acl_get_fd(vfs_handle_struct *handle,
 	return sys_acl_get_fd(handle, fsp, mem_ctx);
 }
 
-static int vfswrap_sys_acl_set_file(vfs_handle_struct *handle,
-				const struct smb_filename *smb_fname,
-				SMB_ACL_TYPE_T acltype,
-				SMB_ACL_T theacl)
+static int vfswrap_sys_acl_set_fd(vfs_handle_struct *handle,
+				  files_struct *fsp,
+				  SMB_ACL_TYPE_T type,
+				  SMB_ACL_T theacl)
 {
-	return sys_acl_set_file(handle, smb_fname, acltype, theacl);
-}
+	if (!fsp->fsp_flags.is_pathref &&
+	    type == SMB_ACL_TYPE_ACCESS)
+	{
+		return sys_acl_set_fd(handle, fsp, theacl);
+	}
 
-static int vfswrap_sys_acl_set_fd(vfs_handle_struct *handle, files_struct *fsp, SMB_ACL_T theacl)
-{
-	return sys_acl_set_fd(handle, fsp, theacl);
+	if (fsp->fsp_flags.have_proc_fds) {
+		int fd = fsp_get_pathref_fd(fsp);
+		struct smb_filename smb_fname;
+		const char *p = NULL;
+		char buf[PATH_MAX];
+
+		p = sys_proc_fd_path(fd, buf, sizeof(buf));
+		if (p == NULL) {
+			return -1;
+		}
+
+		smb_fname = (struct smb_filename) {
+			.base_name = buf,
+		};
+
+		return sys_acl_set_file(handle,
+					&smb_fname,
+					type,
+					theacl);
+	}
+
+	/*
+	 * This is no longer a handle based call.
+	 */
+	return sys_acl_set_file(handle,
+				fsp->fsp_name,
+				type,
+				theacl);
 }
 
 static int vfswrap_sys_acl_delete_def_file(vfs_handle_struct *handle,
@@ -3213,7 +3271,6 @@ struct vfswrap_getxattrat_state {
 	struct tevent_context *ev;
 	files_struct *dir_fsp;
 	const struct smb_filename *smb_fname;
-	struct tevent_req *req;
 
 	/*
 	 * The following variables are talloced off "state" which is protected
@@ -3266,7 +3323,6 @@ static struct tevent_req *vfswrap_getxattrat_send(
 		.ev = ev,
 		.dir_fsp = dir_fsp,
 		.smb_fname = smb_fname,
-		.req = req,
 	};
 
 	max_threads = pthreadpool_tevent_max_threads(dir_fsp->conn->sconn->pool);
@@ -3286,7 +3342,7 @@ static struct tevent_req *vfswrap_getxattrat_send(
 	SMBPROFILE_BYTES_ASYNC_START(syscall_asys_getxattrat, profile_p,
 				     state->profile_bytes, 0);
 
-	if (dir_fsp->fh->fd == -1) {
+	if (fsp_get_pathref_fd(dir_fsp) == -1) {
 		DBG_ERR("Need a valid directory fd\n");
 		tevent_req_error(req, EINVAL);
 		return tevent_req_post(req, ev);
@@ -3424,7 +3480,7 @@ static void vfswrap_getxattrat_do_async(void *private_data)
 		goto end_profile;
 	}
 
-	ret = fchdir(state->dir_fsp->fh->fd);
+	ret = fchdir(fsp_get_pathref_fd(state->dir_fsp));
 	if (ret == -1) {
 		state->xattr_size = -1;
 		state->vfs_aio_state.error = errno;
@@ -3528,9 +3584,34 @@ static ssize_t vfswrap_getxattrat_recv(struct tevent_req *req,
 	return xattr_size;
 }
 
-static ssize_t vfswrap_fgetxattr(struct vfs_handle_struct *handle, struct files_struct *fsp, const char *name, void *value, size_t size)
+static ssize_t vfswrap_fgetxattr(struct vfs_handle_struct *handle,
+				 struct files_struct *fsp,
+				 const char *name,
+				 void *value,
+				 size_t size)
 {
-	return fgetxattr(fsp->fh->fd, name, value, size);
+	int fd = fsp_get_pathref_fd(fsp);
+
+	if (!fsp->fsp_flags.is_pathref) {
+		return fgetxattr(fd, name, value, size);
+	}
+
+	if (fsp->fsp_flags.have_proc_fds) {
+		const char *p = NULL;
+		char buf[PATH_MAX];
+
+		p = sys_proc_fd_path(fd, buf, sizeof(buf));
+		if (p == NULL) {
+			return -1;
+		}
+
+		return getxattr(p, name, value, size);
+	}
+
+	/*
+	 * This is no longer a handle based call.
+	 */
+	return getxattr(fsp->fsp_name->base_name, name, value, size);
 }
 
 static ssize_t vfswrap_listxattr(struct vfs_handle_struct *handle,
@@ -3543,7 +3624,28 @@ static ssize_t vfswrap_listxattr(struct vfs_handle_struct *handle,
 
 static ssize_t vfswrap_flistxattr(struct vfs_handle_struct *handle, struct files_struct *fsp, char *list, size_t size)
 {
-	return flistxattr(fsp->fh->fd, list, size);
+	int fd = fsp_get_pathref_fd(fsp);
+
+	if (!fsp->fsp_flags.is_pathref) {
+		return flistxattr(fd, list, size);
+	}
+
+	if (fsp->fsp_flags.have_proc_fds) {
+		const char *p = NULL;
+		char buf[PATH_MAX];
+
+		p = sys_proc_fd_path(fd, buf, sizeof(buf));
+		if (p == NULL) {
+			return -1;
+		}
+
+		return listxattr(p, list, size);
+	}
+
+	/*
+	 * This is no longer a handle based call.
+	 */
+	return listxattr(fsp->fsp_name->base_name, list, size);
 }
 
 static int vfswrap_removexattr(struct vfs_handle_struct *handle,
@@ -3555,7 +3657,28 @@ static int vfswrap_removexattr(struct vfs_handle_struct *handle,
 
 static int vfswrap_fremovexattr(struct vfs_handle_struct *handle, struct files_struct *fsp, const char *name)
 {
-	return fremovexattr(fsp->fh->fd, name);
+	int fd = fsp_get_pathref_fd(fsp);
+
+	if (!fsp->fsp_flags.is_pathref) {
+		return fremovexattr(fd, name);
+	}
+
+	if (fsp->fsp_flags.have_proc_fds) {
+		const char *p = NULL;
+		char buf[PATH_MAX];
+
+		p = sys_proc_fd_path(fd, buf, sizeof(buf));
+		if (p == NULL) {
+			return -1;
+		}
+
+		return removexattr(p, name);
+	}
+
+	/*
+	 * This is no longer a handle based call.
+	 */
+	return removexattr(fsp->fsp_name->base_name, name);
 }
 
 static int vfswrap_setxattr(struct vfs_handle_struct *handle,
@@ -3570,7 +3693,28 @@ static int vfswrap_setxattr(struct vfs_handle_struct *handle,
 
 static int vfswrap_fsetxattr(struct vfs_handle_struct *handle, struct files_struct *fsp, const char *name, const void *value, size_t size, int flags)
 {
-	return fsetxattr(fsp->fh->fd, name, value, size, flags);
+	int fd = fsp_get_pathref_fd(fsp);
+
+	if (!fsp->fsp_flags.is_pathref) {
+		return fsetxattr(fd, name, value, size, flags);
+	}
+
+	if (fsp->fsp_flags.have_proc_fds) {
+		const char *p = NULL;
+		char buf[PATH_MAX];
+
+		p = sys_proc_fd_path(fd, buf, sizeof(buf));
+		if (p == NULL) {
+			return -1;
+		}
+
+		return setxattr(p, name, value, size, flags);
+	}
+
+	/*
+	 * This is no longer a handle based call.
+	 */
+	return setxattr(fsp->fsp_name->base_name, name, value, size, flags);
 }
 
 static bool vfswrap_aio_force(struct vfs_handle_struct *handle, struct files_struct *fsp)
@@ -3660,7 +3804,6 @@ static struct vfs_fn_pointers vfs_default_fns = {
 
 	/* Directory operations */
 
-	.opendir_fn = vfswrap_opendir,
 	.fdopendir_fn = vfswrap_fdopendir,
 	.readdir_fn = vfswrap_readdir,
 	.readdir_attr_fn = vfswrap_readdir_attr,
@@ -3672,7 +3815,7 @@ static struct vfs_fn_pointers vfs_default_fns = {
 
 	/* File operations */
 
-	.open_fn = vfswrap_open,
+	.openat_fn = vfswrap_openat,
 	.create_file_fn = vfswrap_create_file,
 	.close_fn = vfswrap_close,
 	.pread_fn = vfswrap_pread,
@@ -3724,7 +3867,6 @@ static struct vfs_fn_pointers vfs_default_fns = {
 	.fsctl_fn = vfswrap_fsctl,
 	.set_dos_attributes_fn = vfswrap_set_dos_attributes,
 	.fset_dos_attributes_fn = vfswrap_fset_dos_attributes,
-	.get_dos_attributes_fn = vfswrap_get_dos_attributes,
 	.get_dos_attributes_send_fn = vfswrap_get_dos_attributes_send,
 	.get_dos_attributes_recv_fn = vfswrap_get_dos_attributes_recv,
 	.fget_dos_attributes_fn = vfswrap_fget_dos_attributes,
@@ -3732,13 +3874,13 @@ static struct vfs_fn_pointers vfs_default_fns = {
 	.offload_read_recv_fn = vfswrap_offload_read_recv,
 	.offload_write_send_fn = vfswrap_offload_write_send,
 	.offload_write_recv_fn = vfswrap_offload_write_recv,
-	.get_compression_fn = vfswrap_get_compression,
+	.fget_compression_fn = vfswrap_fget_compression,
 	.set_compression_fn = vfswrap_set_compression,
 
 	/* NT ACL operations. */
 
 	.fget_nt_acl_fn = vfswrap_fget_nt_acl,
-	.get_nt_acl_fn = vfswrap_get_nt_acl,
+	.get_nt_acl_at_fn = vfswrap_get_nt_acl_at,
 	.fset_nt_acl_fn = vfswrap_fset_nt_acl,
 	.audit_file_fn = vfswrap_audit_file,
 
@@ -3748,7 +3890,6 @@ static struct vfs_fn_pointers vfs_default_fns = {
 	.sys_acl_get_fd_fn = vfswrap_sys_acl_get_fd,
 	.sys_acl_blob_get_file_fn = posix_sys_acl_blob_get_file,
 	.sys_acl_blob_get_fd_fn = posix_sys_acl_blob_get_fd,
-	.sys_acl_set_file_fn = vfswrap_sys_acl_set_file,
 	.sys_acl_set_fd_fn = vfswrap_sys_acl_set_fd,
 	.sys_acl_delete_def_file_fn = vfswrap_sys_acl_delete_def_file,
 
